from pickle import loads, dumps

from django.test import TestCase
# from django.db.models import DoesNotExist

from django.contrib.contenttypes.models import ContentType

from stored.queryset import FilterQuerySet
from models import TModel, Rel0
import sys

from stored.boolean import parse, BooleanAlgebra, Symbol


class StoredQuerySetBaseTestCase(TestCase):
    def test_get_model(self):
        self.assert_(FilterQuerySet('Q(a=1)', 3).model == TModel)
        self.assert_(FilterQuerySet('Q(a=1)', 2).model != TModel)
        self.assert_(FilterQuerySet('Q(a=1)', 'stored.tmodel').model == TModel)
        self.assertRaises(ContentType.DoesNotExist, 
                            lambda:FilterQuerySet('Q(a=1)', 'stored.zzzztmodel').model)

    def test_get_manager(self):
        self.assert_(FilterQuerySet('Q(a=1)', 3).manager.model == TModel)
        self.assert_(FilterQuerySet('Q(a=1)', 2).manager.model != TModel)


def flatten(i):
    return [k[0] for k in i]

class FilterQuerySetTestCase(TestCase):
    queries = [
                {'query':"Q(headline__startswith='{headline}')",
                    'selector':'stored.tmodel',
                    'literals':{'headline':'"head1"'},
                    'result':[1, 10],
                },
                {'query':"Q(headline__startswith='{headline}')&Q(id__in={idlist})",
                    'selector':'stored.tmodel',
                    'literals':{'headline':'"head1"', 'idlist':'[1,10]'},
                    'override_literals':{'idlist':'[1,]'},
                    'result':[1, 10],
                    'override_result':[1,]
                },                
                {'query':"Q(related__in=related_qset)",
                    'selector':'stored.tmodel',
                    'eval':{'related_qset':lambda:Rel0.objects.all()[:1]},
                    'result':[1,],
                },
                {'query':"Q(related__in=related_qset)&~Q(id__in=[2,3])",
                    'selector':'stored.tmodel',
                    'eval':{'related_qset':lambda:Rel0.objects.all()[:1]},
                    'result':[1,],
                },
                {'query':"Q(related__in=related_qset)&~Q(id__in={idlist})",
                    'selector':'stored.tmodel',
                    'eval':{'related_qset':lambda:Rel0.objects.all()[:1]},
                    'literals':{'idlist':'[2,3]'},
                    'override_literals':{'idlist':'[1,2,3]'},
                    'result':[1,],
                    'override_result':[],
                },

                ]

    def setUp(self):
        for i in range(10):
            Rel0.objects.create()
        for i in range(1, 11):
            t = TModel.objects.create(headline='head%s'%i)
            t.related.add(Rel0.objects.all()[i-1])
            t.save()

    def test_queries(self):
        for params in self.queries:
            query = FilterQuerySet(params['query'], params['selector'], 
                                        **params.get('literals', {}))
            
            evals = dict((k, v()) for (k,v) in params.get('eval', {}).items())
            evaluated_query = query(**evals)
            result = evaluated_query.values_list('id')

            restored_query = loads(dumps(query))
            restored_result = restored_query(**evals).values_list('id')

            if params.get('override_literals', {}):
                override_result = query.literals(**params['override_literals'])(**evals).values_list('id')
                restored_override_result = restored_query.literals(**params['override_literals'])\
                                            (**evals).values_list('id')
                # self.assert_(query.literals(**params['override_literals']) is not query)

            
            self.assert_(flatten(result) == params['result'],
                         '1:%s:%s!=%s'%(params['query'],
                                        flatten(result),
                                        params['result']))
            self.assert_(flatten(restored_result) == params['result'],
                         '2:%s:%s!=%s'%(params['query'], 
                                        flatten(restored_result),
                                        params['result']))
            self.assert_(flatten(restored_result) == params['result'],
                         '3:%s:%s!=%s'%(params['query'],
                                        flatten(restored_result),
                                        params['result']))
            if params.get('override_literals', {}):
                self.assert_(flatten(override_result) == params['override_result'],
                         '4:%s:%s!=%s'%(params['query'],
                                        flatten(override_result),
                                        params['override_result']))
                self.assert_(flatten(restored_override_result) == params['override_result'],
                         '5:%s:%s!=%s'%(params['query'],
                                        flatten(restored_override_result), 
                                        params['override_result']))

    def test_query_literals(self):
        query = FilterQuerySet("Q(headline__startswith='{headline}')&Q(id__in={idlist})",
                               'stored.tmodel')
        self.assertRaises(KeyError, lambda:query.objects.all())

        query.literals(headline='"head"')
        self.assertRaises(KeyError, lambda:query.objects.all())
        query.literals(idlist='[1,2]')
        self.assert_(flatten(query.objects.all().values_list('id')) == [1,2])

    def test_query_evals(self):
        query = FilterQuerySet("Q(headline__startswith=headline)&Q(id__in={idlist})",
                               'stored.tmodel', idlist='[1,2]')
        self.assertRaises(NameError, lambda:query.objects.all())

        query(headline='head')
        self.assert_(flatten(query.objects.all().values_list('id')) == [1,2])

    def test_query_evals_with_literals0(self):
        query = FilterQuerySet("Q(headline__startswith=headline)&Q(id__in={idlist})",
                               'stored.tmodel')
        self.assertRaises(KeyError, lambda:query.objects.all())

        query.literals(idlist='[1,2]')
        self.assertRaises(NameError, lambda:query.objects.all())

        query(headline='head')
        self.assert_(flatten(query.objects.all().values_list('id')) == [1,2])


    def test_query_evals_with_literals1(self):
        query = FilterQuerySet("Q(headline__startswith=headline)&Q(id__in={idlist})",
                               'stored.tmodel')
        query.literals(idlist='[1,2]')(headline='head')
        self.assert_(flatten(query.objects.all().values_list('id')) == [1,2])

    def test_query_evals_with_literals2(self):
        query = FilterQuerySet("Q(headline__startswith=headline)&Q(id__in={idlist})",
                               'stored.tmodel')
        query(headline='head').literals(idlist='[1,2]')
        self.assert_(flatten(query.objects.all().values_list('id')) == [1,2])

    def test_query_evals_with_literals3(self):
        query = FilterQuerySet("Q(headline__startswith=headline)&Q(id__in={idlist})",
                               'stored.tmodel')
        query(headline='head').literals(idlist='[1,2]')
        query.literals(idlist='[3,4]')
        self.assert_(flatten(query.objects.all().values_list('id')) == [3,4])



from pickle import dumps, loads
from functools import partial, wraps
from ast import literal_eval
from copy import copy

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

def manager_property():
    def fget(self):
        return self._manager

    def fset(self, name):
        self._manager_name = name
        self._manager = copy(getattr(self.model, self._manager_name))

    def fdel(self):
        del self._foo
    return locals()


def default_manager_getter(self):
    return self()

def patch_query_set(stored):
    orig_get_query_set = stored.manager.get_query_set
    if not getattr(orig_get_query_set, 'eval_appent', False):
        @wraps(orig_get_query_set)
        def get_query_set(self, *args, **kwargs):
            return stored.get_query_set(orig_get_query_set(*args, **kwargs))
        get_query_set.eval_appent = True
        stored.manager.get_query_set = get_query_set.__get__(stored.manager,
                                                             stored.manager.__class__)

class FilterQuerySet(object):
    Q = Q
    __slots__ = ['selector',
                 'query',
                 '_literals',
                 '_manager_name',
                 '_manager',
                 '_model',
                 '_eval_locals']
    
    def get_query_set(self, queryset):
        literals = dict((k, literal_eval(v)) for (k, v) in self._literals.items())
        query = eval(self.query.format(**literals), {'Q': FilterQuerySet.Q}, self._eval_locals)
        return queryset.filter(query)

    def literals(self, **literals):
        self._literals.update(literals)
        return self

    def get_model(self):
        if not hasattr(self, '_model'):
            if isinstance(self.selector, int):
                kwargs = {'id': self.selector}
            else:
                kwargs = dict(zip(('app_label', 'model'),
                            self.selector.split('.')))
            self._model = ContentType.objects.get(**kwargs).model_class()
        return self._model

    model = property(get_model)

    manager = property(**manager_property())

    def __init__(self, query, model_selector, manager='objects', **literals):
        if not hasattr(self.__class__, manager):
            setattr(self.__class__, manager, property(default_manager_getter))

        self.selector = model_selector
        self.query = query
        self._literals = literals
        self.manager = manager
        self._eval_locals = {}
        patch_query_set(self)

    def __getstate__(self):
        return dict((k, getattr(self, k)) for k in self.__slots__[:4])

    def __setstate__(self, dict):
        self.__init__(dict['query'],
                      dict['selector'],
                      dict['_manager_name'],
                    **dict['_literals'])

    def __call__(self, **eval_locals):
        self._eval_locals.update(eval_locals)
        return self

    def __getattr__(self, name):
        return getattr(self.manager, name)

StoredQuerySet = FilterQuerySet



#!/usr/bin/env python

from importlib import import_module

import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test


def package_env(file_name, strict=False):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    if os.path.exists(file_path) or strict:
        return open(file_path).read()
    else:
        return u''

PROJECT = u'django-stored-queryset'

VERSION = package_env('VERSION')
URL = package_env('URL')

AUTHOR_AND_EMAIL = [v.strip('>').strip() for v \
                    in package_env('AUTHOR').split('<mailto:')]

if len(AUTHOR_AND_EMAIL) == 2:
    AUTHOR, AUTHOR_EMAIL = AUTHOR_AND_EMAIL
else:
    AUTHOR = AUTHOR_AND_EMAIL
    AUTHOR_EMAIL = u''

DESC = "pickleable Django QuerySet"


class TestRunner(test):
    def run(self, *args, **kwargs):
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(self.\
                                                distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)
        from testrunner import runtests

        runtests()

if __name__ == '__main__':
    setup(
        cmdclass={"test": TestRunner},
        name=PROJECT,
        version=VERSION,
        description=DESC,
        long_description=package_env('README.rst'),
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=package_env('LICENSE'),
        packages=find_packages(),
        package_dir={'stored': 'stored'},
        include_package_data=True,
        zip_safe=True,
        test_suite='test',
        install_requires=['django', ],
        classifiers=[
            'License :: OSI Approved',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Framework :: Django',
        ],
    )
